/* Flex declarations */

%option noyywrap
%option debug
%option caseless
%option 8bit
%option never-interactive

%x STATE_DEFCMD
%x STATE_COMMENT
%x STATE_INCLUDE
%x STATE_DATA

%{
/* C declarations */
#include <cctype>
#include <cstring>

#include "expr.hh"
#include "dimdef.hh"
#include "casevalue.hh"
#include "stringfun.hh"
#include "binding.hh"

/*#include "qb.tab.h" - included by expr.hh */

#define YY_DECL int yylex()

typedef struct saved_yy_context
{
    int line;
    int col;
    char *filename;
    struct saved_yy_context *next;
    YY_BUFFER_STATE state;
    FILE *fp;
} saved_yy_context;
static saved_yy_context *contexts = NULL;
static saved_yy_context current_context;
#define YY_INPUT(buf,result,max_size) \
  { \
    int c; do c = fgetc(yyin); while(c == '\r'); \
    if(c == EOF) result=YY_NULL; \
    else { \
      if(c == '\n')      { ++current_context.line; current_context.col=0; } \
      else if(c == '\t') { current_context.col=(current_context.col+7)&~7; } \
      else { ++current_context.col; } \
      buf[0]=c; result=1; \
    } \
  }

void yyStartFile(FILE *fp, const char *name)
{
    current_context.line     = 1;
    current_context.col      = 0;
    current_context.filename = strdup(name);
    current_context.fp       = fp;
    yyin = fp;
    /* YY_NEW_FILE; */
}
void yyEndFile(void)
{
    free(current_context.filename);
}

void yyerror(const char *msg);
void yyerror(const char *msg)
{
    fprintf(stderr,
        "\n%s:%d:%d: %s\n",
        current_context.filename,
        current_context.line,
        current_context.col,
        msg
    );
}

static char cur_deftype = '!';
static void DefVarType(int begin, int end)
{
    begin = toupper(begin);
    end   = toupper(end);
    if(begin > end) { int c=begin;begin=end;end=c; } /* ensure correct order */
    
    SetDefVarType( (char)begin, (char)end, cur_deftype);
}

%}
/* Flex declarations */

digit [0-9]
blank [ \t]

%%
	/* Grammar begins: */

	/* Exception: DEFTYPE commands */
DEFINT     BEGIN(STATE_DEFCMD); cur_deftype = '%';
DEFSNG     BEGIN(STATE_DEFCMD); cur_deftype = '!';
DEFDBL     BEGIN(STATE_DEFCMD); cur_deftype = '#';
DEFSTR     BEGIN(STATE_DEFCMD); cur_deftype = '$';
DEFLNG     BEGIN(STATE_DEFCMD); cur_deftype = '&';

<STATE_DEFCMD>[A-Z]-[A-Z] DefVarType(yytext[0], yytext[2]);
<STATE_DEFCMD>[A-Z]       DefVarType(yytext[0], yytext[0]);
<STATE_DEFCMD>[, \t]+ /* Ignore comma and space */
<STATE_DEFCMD>:[ :]*  BEGIN(INITIAL); return COLONCHAR; /* end of def */
<STATE_DEFCMD>\n+     BEGIN(INITIAL); return CRLF;      /* end of def */
<STATE_DEFCMD>\'      BEGIN(STATE_COMMENT);


        /* Exception: DATA */
DATA       BEGIN(STATE_DATA); return DATA;
<STATE_DATA>\"[^"\n]*["\n]  {
    fprintf(stderr, "Data string (quoted): '%s'\n", yytext);
    yylval.string = new std::string(yytext);
    return STRINGCONST;
  }
<STATE_DATA>[ \t]+    /* Ignore space in DATA */
<STATE_DATA>[^:,\n]+  {
    while(yyleng > 0 && (*yytext == '\t' || *yytext == ' '))
        { --yyleng; ++yytext; }
    while(yyleng > 0 && (yytext[yyleng-1] == '\t' || yytext[yyleng-1] == ' '))
        { --yyleng; }
    fprintf(stderr, "Data string (unquoted): '%s'\n", yytext);
    /* ltrim and rtrim the string - qb does that too. */
    yylval.string = new std::string(Trim(yytext));
    return STRINGCONST;
  }
<STATE_DATA>,      return *yytext;
<STATE_DATA>:[ :]* BEGIN(INITIAL); return COLONCHAR; /* end of data */
<STATE_DATA>\n+    BEGIN(INITIAL); return CRLF;      /* end of data */



	/* Exception: Comments */
REM\       BEGIN(STATE_COMMENT);
\'         BEGIN(STATE_COMMENT);

<STATE_COMMENT>\$DYNAMIC {
    SetMetaDynamic();
    /* Don't return anything */
  }
<STATE_COMMENT>\$STATIC {
    SetMetaStatic();
    /* Don't return anything */
  }
<STATE_COMMENT>\$INCLUDE:\ *\' BEGIN(STATE_INCLUDE);
<STATE_COMMENT>\n+             BEGIN(INITIAL); return CRLF;
<STATE_COMMENT>[^$\n]+   /* Ignore longest possible blobs until $ */
<STATE_COMMENT>.         /* Ignore other   */

<STATE_INCLUDE>[^']* {
    fprintf(stderr, "Meta $INCLUDE: '%s'\n", yytext);
    {
      current_context.state = YY_CURRENT_BUFFER;
      current_context.fp    = yyin;
      current_context.next  = contexts;
      
      contexts = (saved_yy_context *)malloc(sizeof(*contexts));
      *contexts = current_context;
      
      yyin = fopen(yytext, "rt");
      if(!yyin) { YY_FATAL_ERROR("unable to open include file"); }
      /*setbuf(yyin, NULL);*/
      
      yyStartFile(yyin, yytext);
      
      yy_switch_to_buffer(yy_create_buffer(yyin, YY_BUF_SIZE));
      
      BEGIN(INITIAL);
    }
  }
<STATE_INCLUDE>\'        BEGIN(STATE_COMMENT);

	/* Logical operators */
IMP        return IMP;
EQV        return EQV;
XOR        return XOR;
OR         return  OR;
AND        return AND;
NOT        return NOT;

MOD        return MOD;

	/* If stuff */
IF         return IF;
THEN       return THEN;
ELSE       return ELSE;
ELSEIF     return ELSEIF;
"END IF"   return ENDIF;

	/* Switch stuff */
SELECT     return SELECT;
CASE       return CASE;
"END SELECT"  return ENDSELECT;
IS         return IS;

	/* Loop stuff */
"DO WHILE" return DOWHILE;
"DO UNTIL" return DOUNTIL;
"LOOP WHILE" return LOOPWHILE;
"LOOP UNTIL" return LOOPUNTIL;
DO         return DO;
LOOP       return LOOP;
WHILE      return DOWHILE;
WEND       return LOOP;
"EXIT DO"  return EXITDO;

	/* Sub stuff */
SUB           return SUB;
"END SUB"     return ENDSUB;
"EXIT SUB"    return EXITSUB;
"DECLARE SUB" return DECLARESUB;
FUNCTION           return FUNCTION;
"END FUNCTION"     return ENDFUNCTION;
"EXIT FUNCTION"    return EXITFUNCTION;
"DECLARE FUNCTION" return DECLAREFUNCTION;

	/* For stuff */
FOR        return FOR;
TO         return TO;
STEP       return STEP;
NEXT       return NEXT;
"EXIT FOR" return EXITFOR;

	/* Typedef stuff */
INTEGER    return INTEGER;
SINGLE     return SINGLE;
DOUBLE     return DOUBLE;
STRING     return STRING;
LONG       return LONG;
AS         return AS;
TYPE       return TYPE;
"END TYPE" return ENDTYPE;
SHARED     return SHARED;
DIM        return DIM;
REDIM      return REDIM;
CONST      return CONST;
"OPTION BASE" return OPTION_BASE;
STATIC     return STATIC;

	/* Control stuff */
ERROR      return ERROR;
RESUME     return RESUME;
GOTO       |
"GO TO"    return GOTO;
GOSUB      return GOSUB;
RETURN     return RETURN;
RESTORE    return RESTORE;
READ       return READ;
CALL       return CALL;
END        return END;
	
	/* Something */
"DEF SEG"  return DEFSEG;

	/* File stuff */
OPEN       return OPEN;
OUTPUT     return OUTPUT;
INPUT      return INPUT;
APPEND     return APPEND;
RANDOM     return RANDOM;
BINARY     return BINARY;
WRITE      return WRITE;
ACCESS     return ACCESS;
LOCK       return LOCK;
CLOSE      return CLOSE;

	/* Other I/O stuff */
\?\ *\#    |
PRINT\ +\# return PRINT_F;
\?         |
PRINT      return PRINT;
USING      return USING;
LPRINT     return LPRINT;

	/* Graphics */
LOCATE     return LOCATE;
GET        return GET;
PUT        return PUT;
SCREEN     return SCREEN;
PALETTE    return PALETTE;
VIEW       return VIEW;
PSET       return PSET;
PRESET     return PRESET;
LINE       return LINE;
CIRCLE     return CIRCLE;
PAINT      return PAINT;
DRAW       return DRAW;
COLOR      return COLOR;
CLS        return CLS;

	/* Other */
LET        return LET;
ON         return ON;
OFF        return OFF;
"ON PLAY"  return ON_PLAY;
PLAY       return PLAY;
STOP       return STOP;
POKE       return POKE;
SYSTEM     return SYSTEM;
OUT        return OUT;
WAIT       return WAIT;

	/* Functions without parameters*/
INKEY\$    return INKEY_S;
CSRLIN     return CSRLIN;
FREEFILE   return FREEFILE;
DATE\$     return DATE_S;
TIME\$     return TIME_S;
TIMER      return TIMER;
ERL        return ERL;
ERR        return ERR;
RND        return RND;
        /* Functions taking strings */
ASC        return ASC;
LEN        return LEN;
CVD        return CVD;
CVI        return CVI;
CVL        return CVL;
CVS        return CVS;
LTRIM\$    return LTRIM_S;
RTRIM\$    return RTRIM_S;
LCASE\$    return LCASE_S;
UCASE\$    return UCASE_S;
VAL        return VAL;
        /* Functions with mixed parameters */
ENVIRON\$  return ENVIRON_S; /* also takes strings */
FILEATTR   return FILEATTR; /* two args */
FRE        return FRE;       /* also takes strings */
INSTR      return INSTR;
MID\$      return MID_S;
LEFT\$     return LEFT_S;
RIGHT\$    return RIGHT_S;
        /* Functions taking numbers */
ABS        return ABS;
ATN        return ATN;
CDBL       return CDBL;
CINT       return CINT;
CHR\$      return CHR_S;
CLNG       return CLNG;
COS        return COS;
CSNG       return CSNG;
EXP        return EXP;
FIX        return FIX;
HEX\$      return HEX_S;
INP        return INP;
INPUT\$    return INPUT_S;
INT        return INT;
LOC        return LOC;
LOF        return LOF;
LOG        return LOG;
LPOS       return LPOS;
MKD\$      return MKD_S;
MKI\$      return MKI_S;
MKL\$      return MKL_S;
MKS\$      return MKS_S;
OCT\$      return OCT_S;
POINT      return POINT; /* gfx */
PEEK       return PEEK;
POS        return POS;
SPC        return SPC;
SQR        return SQR;
STR\$      return STR_S;
STRING\$   return STRING_S;
TAB        return TAB;
TAN        return TAN;
SPACE\$    return SPACE_S;
        /* Meta-functions */
LBOUND     return LBOUND; /* array */
UBOUND     return UBOUND; /* array */
VARPTR     return VARPTR;   /* any */
VARPTR\$   return VARPTR_S; /* any */
VARSEG     return VARSEG;   /* any */


[a-z][a-z0-9]*[!#%$&]? {
    fprintf(stderr, "Identifier: '%s' (%u)\n", yytext, yyleng);
    yylval.string = new std::string(yytext);
    return IDENTIFIER;
  }

-?{digit}+[%&] {
    fprintf(stderr, "Integer: '%s'\n", yytext);
    yylval.expr = new expr_number(strtoll(yytext,0,10));
    return NUMBER;
  }

&H[0-9A-F]+[%&]? {
    fprintf(stderr, "Hex integer: '%s'\n", yytext);
    yylval.expr = new expr_number(strtoll(yytext+2,0,16));
    return NUMBER;
  }

-?{digit}+(\.{digit}*)?(e[+-]?{digit}+)?[!#]? {
    yylval.expr = new expr_number(strtold(yytext,0));
    return NUMBER;
  }

-?\.{digit}+(e[+-]?{digit}+)?[!#]? {
    yylval.expr = new expr_number(strtold(yytext,0));
    return NUMBER;
  }

#{digit}* {
    ++yytext; --yyleng;
    yylval.expr = new expr_number(strtoll(yytext,0,10));
    return FILENUM;
  }

:[ :]*  BEGIN(INITIAL); return COLONCHAR;
\n+     BEGIN(INITIAL); return CRLF;

"<="    return LESSOREQUAL;
">="    return GREATEROREQUAL;
"<>"    |
"><"    |
"!="    return NOTEQUAL;

\"[^"\n]*\"  {
    ++yytext; yyleng -= 2;
    fprintf(stderr, "String: '%s'\n", yytext);
    yylval.string = new std::string(yytext);
    return STRINGCONST;
  }

[-<>=+*/\\,;().^]   return *yytext;

{blank}+    /* blank   */

<<EOF>> {
    fclose(yyin);
    if(contexts)
    {
      yyEndFile();
      current_context = *contexts;
      yyin = current_context.fp;
      free(contexts);
      contexts = current_context.next;
      
      yy_delete_buffer(YY_CURRENT_BUFFER);
      yy_switch_to_buffer(current_context.state);
    }
    else
    {
      yyterminate();
    }
  }

<*>.    BEGIN(INITIAL); return UNKNOWN_CHARACTER;

%%
